"""
Batch simulation runner
"""

from typing import Optional, Union

import chex
import jax

from .sim import Sim, TSimParams
from .utils.functions import get_size


def batch_sim_runner(
    sim: Sim,
    n_samples: int,
    n_steps: int,
    rng: Union[chex.PRNGKey, int],
    show_progress: bool = True,
    params: Optional[TSimParams] = None,
    param_samples: Optional[TSimParams] = None,
    **step_kwargs,
) -> chex.ArrayTree:
    """
    Batch Monte-Carlo and parameter sweep simulation execution

    Run simulations across multiple random keys and optionally
    across a sample of simulation parameters. If run across
    parameters, then in will run ``n_samples`` per parameter
    sample.

    Parameters
    ----------
    sim
        Simulation functionality class.
    n_steps
        Number of simulation steps.
    n_samples
        Number of Monte-Carlo samples.
    rng
        Either a JAX random key or integer random seed.
    show_progress
        If ``True`` simulation progress bar will be displayed.
        Default ``True``.
    params
        Optional simulation parameters. If not provided default
        simulation parameters will be used.
    param_samples
        Optional simulation parameter samples to generate
        data across. Should have the same tree structure as individual
        simulation parameters.
    **step_kwargs
        Any additional keyword arguments passed to the
        step function. Arguments are static over the
        course of the simulation.

    Returns
    -------
    Any
        Simulation records, with shape ``[n_samples, n_steps]`` if run
        with a single set of parameters, or shape ``[n_params, n_samples, n_steps]``
        if a set of parameters is provided.
    """
    key = rng if isinstance(rng, chex.PRNGKey) else jax.random.PRNGKey(rng)

    assert (
        params is None or param_samples is None
    ), "Either params or param_samples should be provided, not both"

    if params is None and param_samples is None:
        params = sim.default_params()

    def inner(k, i, _params):
        _, records = sim.init_and_run(
            n_steps,
            k,
            show_progress=show_progress,
            pbar_id=i,
            params=_params,
            **step_kwargs,
        )
        return records

    def sample_params(k, _params, pbar_offset):
        pbar_offset = pbar_offset * n_samples
        keys = jax.random.split(k, n_samples)
        return jax.vmap(inner, in_axes=(0, 0, None))(
            keys, pbar_offset + jax.numpy.arange(n_samples), _params
        )

    if params is not None:
        batch_records = sample_params(key, params, 0)
    else:
        n_params = get_size(param_samples)
        batch_records = jax.vmap(sample_params, in_axes=(None, 0, 0))(
            key, param_samples, jax.numpy.arange(n_params)
        )

    return batch_records



from typing import Tuple

import chex
import jax
import jax.numpy as jnp
import optax
from flax import linen as nn
from flax.training.train_state import TrainState
from typing_extensions import Self


class AgentState(TrainState):
    """
    Basic RL agent parameter and optimiser state

    Extends :py:class:`flax.training.train_state.TrainState`,
    tracks the current network parameters and optimiser state.
    """

    @classmethod
    def init_from_model(
        cls,
        key: chex.PRNGKey,
        model: nn.Module,
        tx: optax.GradientTransformation,
        observation_shape: Tuple[int, ...],
    ) -> Self:
        """
        Initialise state from a model/policy

        Parameters
        ----------
        key
            JAX random key.
        model
            Flax neural network model definition.
        tx
            Optax optimiser.
        observation_shape
            Shape of observations

        Returns
        -------
        esquilax.ml.rl.AgentState
            Initialised agent state.
        """
        fake_args = jnp.zeros(observation_shape)
        params = model.init(key, fake_args)
        return cls.create(apply_fn=model.apply, params=params, tx=tx)

    def apply(self, observations: chex.Array) -> chex.Array:
        """
        Apply the network to a batch of observations

        Parameters
        ----------
        observations
            Array of observations, where the first
            axis of the array should correspond to the
            number of agents.

        Returns
        -------
        chex.Array
            Array of network output for each agent.
        """
        return jax.vmap(self.apply_fn, in_axes=(None, 0))(self.params, observations)


class BatchAgentState(TrainState):
    """
    Batch agent state

    Agent state for use where each agent has individual parameters
    and optimiser state (but the same network structure and optimiser).
    """

    @classmethod
    def init_from_model(
        cls,
        key: chex.PRNGKey,
        model: nn.Module,
        tx: optax.GradientTransformation,
        observation_shape: Tuple[int, ...],
        n_agents: int,
    ) -> Self:
        """
        Initialise state from a network

        Will generate individual (random) initial
        parameters for each agent.

        Parameters
        ----------
        key
            JAX random key.
        model
            Flax neural network model definition.
        tx
            Optax optimiser.
        observation_shape
            Shape of observations
        n_agents
            Number of agents to initialise state for.

        Returns
        -------
        esquilax.ml.rl.BatchAgentState
            Initialised agent-states.
        """
        fake_args = jnp.zeros(observation_shape)

        def init(_k):
            params = model.init(_k, fake_args)
            return cls.create(apply_fn=model.apply, params=params, tx=tx)

        keys = jax.random.split(key, n_agents)
        return jax.vmap(init)(keys)

    def apply_gradients(self, *, grads, **kwargs) -> Self:
        """
        Apply gradient updates to agent parameters

        Map apply a batch of gradient (i.e. individual updates per agent)
        across the individual agent parameters.

        Parameters
        ----------
        grads
            Batch of gradients, the first axis of the updates
            should correspond to the individual agents.
        **kwargs
            Any additional keyword arguments to forward to
            the underlying ``apply_gradients`` method.

        Returns
        -------
        esquilax.ml.rl.BatchAgentState
            Updated state
        """
        return jax.vmap(lambda a, g: TrainState.apply_gradients(a, grads=g, **kwargs))(
            self, grads
        )

    def apply(self, observations: chex.Array):
        """
        Apply the network, using individual agent parameters

        Applies the network mapped across agent parameters
        and corresponding observations

        Parameters
        ----------
        observations
            Array of observations, where the first
            axis of the array should correspond to the
            number of agents.

        Returns
        -------
        chex.Array
            Array of network output for each agent.
        """
        return jax.vmap(self.apply_fn, in_axes=(0, 0))(self.params, observations)



"""
Reinforcement learning utilities and training functionality
"""
from .agent import Agent
from .agent_state import AgentState, BatchAgentState
from .environment import Environment
from .training import test, train, train_and_test
from .types import Trajectory



"""
Generic types
"""
from typing import Collection, TypeVar

import chex

TSimState = TypeVar("TSimState")
"""Generic simulation state"""
TSimParams = TypeVar("TSimParams")
"""Generic simulation parameters"""
TEnvState = TypeVar("TEnvState")
"""Generic environment state"""
TEnvParams = TypeVar("TEnvParams")
"""Generic environment parameters"""
T = TypeVar("T")
TypedPyTree = T | Collection[T]
"""Reduction function(s) type"""
Default = bool | int | float | chex.ArrayTree
"""Default reduction value types"""
